/*****************************************************************************/
/*                      WiNRADiO G65DDC EXCALIBUR SIGMA                      */
/*                       Coherent mode SDK DEMO - ETC                        */
/*                        (c) Copyright RADIXON 2023                         */
/*****************************************************************************/

#include <stdio.h>
#include <vector>
#include "u_rx_server.h"
#include <map>
#include <string>
#include <dlfcn.h>
#include <unistd.h>
#include <string.h>
#include <wx/wx.h>
#include <wx/fileconf.h>

#define DEMO_RX
#define DEFAULT_PORT 7000

void *hapi;
G65DDC_CREATE_INSTANCE create_instance;

bool load_api(void)
{
    hapi=dlopen("libg65ddcapi.so",RTLD_LAZY);

    if(hapi==NULL)
    {
        printf("Failed to load libg65ddcapi.so\n");
        return false;
    }

    create_instance=(G65DDC_CREATE_INSTANCE)dlsym(hapi,"CreateInstance");

    if(create_instance==NULL)
    {
        printf("Failed to load valid libg65ddcapi.so!");
        dlclose(hapi);
        return false;
    }

    //test for required version of the API which supports extended configuration
    if(dlsym(hapi,"CohSetExtendedConfigurationParams")==NULL)
    {
        printf("Failed to load proper version of the libg65ddcapi.so.\n");
        printf("Please install the lastest version of the G65DDC software.\n");
        return false;
    }

    return true;
}

//open each single receiver as separate coherent device set
bool open_receivers(std::vector<ICohG65DDCDeviceSet*> &device_sets)
{
 IG65DDCDeviceEnumerator *enumerator;
 uint32_t i,count;
 G65DDC_DEVICE_INFO device_info;
 ICohG65DDCDeviceSet *device_set;

    device_sets.clear();

    if(!create_instance(G65DDC_CLASS_ID_DEVICE_ENUMERATOR,(void**)&enumerator))
    {
        printf("Failed to craete device enumerator object, error=%u.\n",errno);
        return false;
    }

    count=enumerator->GetCount();

    for(i=0;i<count;i++)
    {
        enumerator->GetDeviceInfo(i,&device_info);

        if(device_info.Flags & G65DDC_FLAGS_COHERENT)
        {
            if(!create_instance(G65DDC_CLASS_ID_COH_DEVICE_SET,(void**)&device_set))
            {
                printf("Failed to create device set object, error=%u.\n",errno);
                enumerator->Free();
                return false;
            }

            if(!device_set->Open(&device_info,1))
            {
                printf("Failed to open device %s, error=%u.\n",device_info.SerialNumber,errno);
                device_set->Free();
                enumerator->Free();
                return false;
            }

            device_sets.push_back(device_set);
        }
    }

    enumerator->Free();

    if(device_sets.empty())
    {
        printf("No coherent G65DDC receiver has been found!\n");
#ifndef DEMO_RX
        return false;
#else
        printf("DEMO MODE - 3 demo receivers\n");

        for(int i=0;i<3;i++)
        {
            create_instance(G65DDC_CLASS_ID_COH_DEVICE_SET,(void**)&device_set);

            device_set->Open(COH_G65DDC_OPEN_DEMO,1);
            device_sets.push_back(device_set);
        }
#endif
    }
    else
    {
        printf("Number of coherent G65DDC receivers: %u\n",(uint32_t)device_sets.size());
    }

    return true;
}

void read_config(std::map<std::string,uint16_t> &config)
{
 wxFileConfig config_file(wxEmptyString,wxEmptyString,wxGetCwd()+wxT("/G65DDCCohETCDemo_Server.ini"));
 wxString s;
 int port;
 wxString group;
 long index;

    config.clear();

    config_file.SetPath(wxT("/"));

    if(config_file.GetFirstGroup(group,index))
    {
        do
        {
            config_file.SetPath(group);

            port=config_file.ReadLong(wxT("Port"),DEFAULT_PORT);

            config[group.mb_str().data()]=(uint16_t)port;

            config_file.SetPath(wxT("/"));
        }
        while(config_file.GetNextGroup(group,index));
    }
}

int main(void)
{
 std::vector<ICohG65DDCDeviceSet*> device_sets;
 std::vector<rx_server*> servers;
 std::map<std::string,uint16_t> config;
 uint16_t default_port,port;
 char serial_number[9];

    wxInitializer initializer;

    if(!initializer)
    {
        printf("Failed to initialize wxWidget library!\nPress ENTER to exit.\n");
        getchar();
        return 0;
    }

    if(!load_api())
    {
        printf("\nPress ENTER to exit.\n");
        getchar();
        return 0;
    }

    if(!open_receivers(device_sets))
    {
        printf("\nPress ENTER to exit.\n");
        getchar();
        return 0;
    }

    read_config(config);

    default_port=DEFAULT_PORT;

    for(auto device_set:device_sets)
    {
        auto server=new rx_server(device_set);

        strcpy(serial_number,server->get_serial_number());

        for(int i=0;serial_number[i];i++)
        {
            if(serial_number[i]<=' ')
            {
                serial_number[i]='\0';
                break;
            }
        }

        auto it=config.find(serial_number);

        if(it!=config.end())
        {
            port=it->second++;
        }
        else
        {
            port=default_port++;
        }

        if(!server->start(port))
        {
            break;
        }

        servers.push_back(server);
    }

    printf("\nPress ENTER to exit.\n");
    getchar();

    printf("Exiting....");


    for(auto server:servers)
    {
        delete server;
    }

    printf("\r               ");

    return 0;
}
