/*****************************************************************************/
/*                      WiNRADiO G65DDC EXCALIBUR SIGMA                      */
/*                       Coherent mode SDK DEMO - ETC                        */
/*                        (c) Copyright RADIXON 2023                         */
/*****************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include "u_rx_server.h"
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <functional>
#include <string.h>


void dbg_print(const char *format,...)
{
 va_list list;

    va_start(list,format);

    vprintf(format,list);

    va_end(list);
}

rx_server::rx_server(ICohG65DDCDeviceSet *device_set)
{
    m_device_set=device_set;
    m_listen_socket=-1;
    m_client_socket=-1;

    m_packet_seq=0;

    device_set->GetDeviceInfo(0,&m_device_info);

    device_set->SetConfiguration(COH_G65DDC_CONFIGURATION_EXTENDED);

    device_set->SetCallback(this);

    m_packet_ddc1_samples=(union tcp_packet*)malloc(offsetof(tcp_packet,s_ddc1_samples.data)+6*MAX_DDC1_SAMPLE_SET_COUNT);
}

rx_server::~rx_server(void)
{
    stop();

    free(m_packet_ddc1_samples);
}


bool rx_server::start(uint16_t port)
{
 struct sockaddr_in addr;
 int enable;

    printf("Starting server for %s",m_device_info.SerialNumber);

    m_listen_socket=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

    if(m_listen_socket==-1)
    {
        printf("\nFailed to create socket, error=%d\n",errno);
        return false;
    }

    enable=1;
    setsockopt(m_listen_socket,SOL_SOCKET,SO_REUSEADDR,&enable,sizeof(enable));

    addr.sin_family=AF_INET;
    addr.sin_addr.s_addr=htonl(INADDR_ANY);
    addr.sin_port=htons(port);

    if(bind(m_listen_socket,(sockaddr*)&addr,sizeof(addr))==-1)
    {
        printf("\nFailed to bind socket, error=%d\n",errno);
        close(m_listen_socket);
        m_listen_socket=-1;
        return false;
    }

    listen(m_listen_socket,5);

    m_stop_request=false;

    m_thread=new std::thread(std::bind(&rx_server::working_thread_proc,this));

    printf("\r%s:%u                                              \n",m_device_info.SerialNumber,port);

    return true;
}

void rx_server::stop(void)
{
    if(m_listen_socket==-1)
    {
        return;
    }

    m_stop_request=true;
    m_thread->join();
    delete m_thread;
    m_thread=NULL;

    close(m_listen_socket);
    m_listen_socket=-1;
}

void rx_server::CohG65DDC_DDC1StreamCallback(ICohG65DDCDeviceSet *DeviceSet,uint32_t DeviceCount,const void **Buffers,uint32_t NumberOfSamples,uint32_t BitsPerSample)
{
 uint32_t count;
 uint8_t *src,*dst;

    count=m_ddc1_sample_sets_to_send;

    m_packet_ddc1_samples->code=PACKET_CODE_S_DDC1_SAMPLES;
    m_packet_ddc1_samples->s_ddc1_samples.index=m_ddc1_sample_set_counter;
    m_packet_ddc1_samples->s_ddc1_samples.count=count;
    m_packet_ddc1_samples->s_ddc1_samples.bits_per_sample=BitsPerSample;

    m_ddc1_sample_set_counter+=NumberOfSamples;

    if(BitsPerSample==16)
    {
        m_packet_ddc1_samples->size=PACKET_SIZE(s_ddc1_samples)-1+count*4;
        memcpy(m_packet_ddc1_samples->s_ddc1_samples.data,Buffers[0],count*4);
    }
    else
    {
        m_packet_ddc1_samples->size=PACKET_SIZE(s_ddc1_samples)-1+count*6;
        src=(uint8_t*)Buffers[0];
        dst=m_packet_ddc1_samples->s_ddc1_samples.data;

        for(uint32_t i=0;i<count;i++)
        {
            dst[i*6+0]=src[i*8+1];
            dst[i*6+1]=src[i*8+2];
            dst[i*6+2]=src[i*8+3];

            dst[i*6+3]=src[i*8+5];
            dst[i*6+4]=src[i*8+6];
            dst[i*6+5]=src[i*8+7];
        }
    }

    send(*m_packet_ddc1_samples);
}

void rx_server::CohG65DDC_ADCSnapshotCallback(ICohG65DDCDeviceSet *DeviceSet,uint32_t DeviceIndex,const short *Buffer,uint32_t Count,uint32_t CenterFrequency,uint16_t ADCLevel)
{
}

void  rx_server::CohG65DDC_DDC2StreamCallback(ICohG65DDCDeviceSet *DeviceSet,uint32_t DeviceIndex,const float *Buffer,uint32_t NumberOfSamples)
{
}

void  rx_server::CohG65DDC_DDC2PreprocessedStreamCallback(ICohG65DDCDeviceSet *Device,uint32_t DeviceIndex,const float *Buffer,uint32_t NumberOfSamples,float SlevelPeak,float SlevelRMS)
{
}

void  rx_server::CohG65DDC_AudioStreamCallback(ICohG65DDCDeviceSet *Device,uint32_t DeviceIndex,uint32_t Type,const float *Buffer,uint32_t NumberOfSamples)
{
}

void rx_server::working_thread_proc(void)
{
 fd_set read_fds;
 int result;
 timeval timeout;
 int socket;
 int bytes_in_queue;
 int nfds;
 union tcp_packet *packet;
 uint32_t bytes_received;
 uint32_t bytes_to_recv;
 uint32_t total_packet_size;

    timeout.tv_sec=0;
    timeout.tv_usec=250000;

    packet=(union tcp_packet*)malloc(MAX_PACKET_SIZE);
    bytes_received=0;


    auto interval_start = std::chrono::steady_clock::now();

    while(!m_stop_request)
    {
        FD_ZERO(&read_fds);
        FD_SET(m_listen_socket,&read_fds);

        nfds=m_listen_socket;

        if(m_client_socket!=-1)
        {
            FD_SET(m_client_socket,&read_fds);

            if(nfds<m_client_socket)
            {
                nfds=m_client_socket;
            }

            auto interval_end=std::chrono::steady_clock::now();

            if(std::chrono::duration_cast<std::chrono::milliseconds>(interval_end-interval_start).count()>=250)
            {
                send_state();
                interval_start=interval_end;
            }
        }

        result=select(nfds+1,&read_fds,NULL,NULL,&timeout);

        if(result>0)
        {
            if(FD_ISSET(m_listen_socket,&read_fds))
            {
                struct sockaddr_in addr;
                socklen_t addr_len=sizeof(addr);

                socket=accept(m_listen_socket,(sockaddr*)&addr,&addr_len);

                if(m_client_socket==-1)
                {
                    //no client is connected
                    sprintf(m_client_address_str,"%s:%u",inet_ntoa(addr.sin_addr),ntohs(addr.sin_port));
                    DEBUG("%s:conenction accepted",m_client_address_str);

                    m_client_socket=socket;
                    bytes_received=0;

                    send_device_info();
                }
                else
                {
                    //a client is already connected
                    close(socket);
                    DEBUG("conenction from %s:%u rejected:server is busy",inet_ntoa(addr.sin_addr),ntohs(addr.sin_port));
                }
            }

            if(m_client_socket!=-1 && FD_ISSET(m_client_socket,&read_fds))
            {
                if(ioctl(m_client_socket,FIONREAD,&bytes_in_queue)==0)
                {
                    if(bytes_in_queue>0)
                    {
                        do
                        {
                            if(bytes_received>=EMPTY_PACKET_SIZE)
                            {
                                total_packet_size=sizeof(packet_size_t)+packet->size;

                                bytes_to_recv=total_packet_size-bytes_received;

                                if(bytes_to_recv>(uint32_t)bytes_in_queue)
                                {
                                    bytes_to_recv=bytes_in_queue;
                                }

                                result=recv(m_client_socket,packet->c+bytes_received,bytes_to_recv,0);

                                if(result>0)
                                {
                                    bytes_received+=result;
                                    bytes_in_queue-=result;

                                    if(total_packet_size==bytes_received)
                                    {
                                        process_packet(packet);
                                        bytes_received=0;
                                    }
                                }
                                else if(result==0)
                                {
                                    //disconnected
                                    close_client_socket();
                                    break;
                                }
                            }
                            else
                            {
                                bytes_to_recv=EMPTY_PACKET_SIZE-bytes_received;

                                if(bytes_to_recv>(uint32_t)bytes_in_queue)
                                {
                                    bytes_to_recv=bytes_in_queue;
                                }

                                result=recv(m_client_socket,packet->c+bytes_received,bytes_to_recv,0);

                                if(result>0)
                                {
                                    bytes_received+=result;
                                    bytes_in_queue-=result;
                                }
                                else if(result==0)
                                {
                                    //disconnected
                                    close_client_socket();
                                    break;
                                }
                            }
                        }
                        while(bytes_in_queue>0);
                    }
                    else
                    {
                        //disconnected
                        close_client_socket();
                    }
                }
            }
        }
    }

    if(m_client_socket!=-1)
    {
        close_client_socket();
    }

    free(packet);
}

void rx_server::close_client_socket(void)
{
    m_device_set->SetPower(0);

    close(m_client_socket);
    m_client_socket=-1;

    DEBUG("%s:connection closed",m_client_address_str);
}


bool rx_server::send(union tcp_packet &packet)
{
 int result;
 uint32_t bytes;
 void *buffer;

    bytes=packet.size+sizeof(packet_size_t);
    buffer=packet.u8;

    m_send_cs.lock();

    packet.seq=m_packet_seq++;

    while(bytes)
    {
        result=::send(m_client_socket,(char*)buffer,bytes,0);

        if(result<0)
        {
            m_send_cs.unlock();
            return false;
        }

        bytes-=result;
        buffer=(char*)buffer+result;
    }

    m_send_cs.unlock();

    return true;
}

void rx_server::send_device_info(void)
{
 union tcp_packet packet;

    packet.code=PACKET_CODE_S_DEVICE_INFO;
    packet.size=PACKET_SIZE(s_device_info);

    strncpy(packet.s_device_info.serial_number,m_device_info.SerialNumber,sizeof(packet.s_device_info.serial_number));

    send(packet);
}

void rx_server::send_state(void)
{
 union tcp_packet packet;

    m_device_set->GetDeviceState(0,&packet.s_state.device);

    if(!m_device_set->Get1PPSCounters(&packet.s_state.event_counter,&packet.s_state.adc_period_counter))
    {
        packet.s_state.event_counter=-1;
    }

    if(!m_device_set->Get1PPSDDC1Counters(&packet.s_state.sample_counter,&packet.s_state.adc_counter))
    {
        packet.s_state.sample_counter=-1.0;
    }

    packet.code=PACKET_CODE_S_STATE;
    packet.size=PACKET_SIZE(s_state);

    send(packet);
}


uint32_t rx_server::process_packet_set_extended_configuration_params(union tcp_packet *packet)
{
    DEBUG("%s:extended configuration params:index=%u:device count=%u:delay=%d",m_client_address_str,packet->c_extended_configuration_params.index,
        packet->c_extended_configuration_params.device_count,packet->c_extended_configuration_params.delay);

    if(m_device_set->SetExtendedConfigurationParams(packet->c_extended_configuration_params.index,
        packet->c_extended_configuration_params.device_count))
    {
        if(m_device_set->SetTriggerDelay(packet->c_extended_configuration_params.delay))
        {
            return 0;
        }
    }

    return errno;
}

uint32_t rx_server::process_packet_set_power(union tcp_packet *packet)
{
    DEBUG("%s:set power:%s",m_client_address_str,packet->c_power.power?"ON":"OFF");

    if(m_device_set->SetPower(packet->c_power.power))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_start_ddc1(union tcp_packet *packet)
{
    DEBUG("%s:start ddc1:index=%u:sample sets per buffer=%u:device set=%p",m_client_address_str,packet->c_start_ddc1.type_index,
            packet->c_start_ddc1.sample_sets_per_buffer,m_device_set);

    m_device_set->StopDDC1();
    m_device_set->SetDDC1(packet->c_start_ddc1.type_index);

    m_ddc1_sample_set_counter=0;
    m_ddc1_sample_sets_to_send=packet->c_start_ddc1.sample_sets_to_send;

    if(m_device_set->StartDDC1(packet->c_start_ddc1.sample_sets_per_buffer))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_stop_ddc1(void)
{
    DEBUG("%s:stop ddc1:device set=%p",m_client_address_str,m_device_set);

    m_device_set->StopDDC1();

    return 0;
}

uint32_t rx_server::process_packet_trigger(void)
{
    DEBUG("%s:trigger",m_client_address_str);

    if(m_device_set->Trigger())
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_trigger_test(void)
{
    DEBUG("%s:trigger test",m_client_address_str);

    if(m_device_set->TriggerTest())
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_set_ddc1_frequency(union tcp_packet *packet)
{
 uint32_t min_frequency,max_frequency,frequency;
 uint32_t range;
 int32_t diff_1,diff_2;
 G65DDC_DDC_INFO ddc_info;

    DEBUG("%s:set ddc1 frequency to %d Hz",m_client_address_str,packet->c_ddc1_frequency.frequency);

    frequency=packet->c_ddc1_frequency.frequency;

    m_device_set->GetDDC1(NULL,&ddc_info);

    min_frequency=m_device_info.Ranges[G65DDC_RANGE_1].MinFrequency+ddc_info.Bandwidth/2;
    max_frequency=m_device_info.Ranges[G65DDC_RANGE_2].MaxFrequency-ddc_info.Bandwidth/2;

    if(frequency<min_frequency)
    {
        frequency=min_frequency;
    }
    else if(frequency>max_frequency)
    {
        frequency=max_frequency;
    }

    diff_1=m_device_info.Ranges[G65DDC_RANGE_1].MaxFrequency-frequency;
    diff_2=frequency-m_device_info.Ranges[G65DDC_RANGE_2].MinFrequency;

    if(diff_1>0)
    {
        //frequency is in range1
        range=G65DDC_RANGE_1;
    }
    else if(diff_2>0)
    {
        //Frequency is in range2
        range=G65DDC_RANGE_2;
    }
    else
    {
        //Frequency is between ranges, select the closest range
        if(abs(diff_1)<abs(diff_2))
        {
            range=G65DDC_RANGE_1;
            frequency=m_device_info.Ranges[G65DDC_RANGE_1].MaxFrequency-ddc_info.Bandwidth/2;
        }
        else
        {
            range=G65DDC_RANGE_2;
            frequency=m_device_info.Ranges[G65DDC_RANGE_2].MinFrequency+ddc_info.Bandwidth/2;
        }
    }

    if(!m_device_set->SetRange(range))
    {
        return errno;
    }

    if(m_device_set->SetDDC1Frequency(frequency))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_set_ddc1_phase_shift(union tcp_packet *packet)
{
    DEBUG("%s:set ddc1 phase shift to %.6f",m_client_address_str,packet->c_ddc1_phase_shift.phase_shift);

    if(m_device_set->SetDDC1PhaseShift(0,packet->c_ddc1_phase_shift.phase_shift))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_set_atten(union tcp_packet *packet)
{
    DEBUG("%s:set attenuator to %u dB",m_client_address_str,packet->c_atten.atten);

    if(m_device_set->SetAttenuator(packet->c_atten.atten))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_set_preamp(union tcp_packet *packet)
{
    DEBUG("%s:set preamp to %u dB",m_client_address_str,packet->c_preamp.preamp);

    if(m_device_set->SetPreamplifier(packet->c_preamp.preamp))
    {
        return 0;
    }

    return errno;
}

uint32_t rx_server::process_packet_set_1pps(union tcp_packet *packet)
{
    DEBUG("%s:set 1pps to %u",m_client_address_str,packet->c_1pps.enabled);

    if(m_device_set->Set1PPSEnabled(packet->c_1pps.enabled))
    {
        return 0;
    }

    return errno;
}

void rx_server::process_packet(union tcp_packet *packet)
{
 uint32_t error;
 union tcp_packet ack_packet;

    error=0;

    switch(packet->code)
    {
        case PACKET_CODE_C_SET_DDC1_FREQUENCY:
            error=process_packet_set_ddc1_frequency(packet);
            break;

        case PACKET_CODE_C_SET_DDC1_PHASE_SHIFT:
            error=process_packet_set_ddc1_phase_shift(packet);
            break;

        case PACKET_CODE_C_START_DDC1:
            error=process_packet_start_ddc1(packet);
            break;

        case PACKET_CODE_C_STOP_DDC1:
            error=process_packet_stop_ddc1();
            break;

        case PACKET_CODE_C_TRIGGER:
            error=process_packet_trigger();
            break;

        case PACKET_CODE_C_TRIGGER_TEST:
            error=process_packet_trigger_test();
            break;

        case PACKET_CODE_C_SET_ATTEN:
            error=process_packet_set_atten(packet);
            break;

        case PACKET_CODE_C_SET_PREAMP:
            error=process_packet_set_preamp(packet);
            break;

        case PACKET_CODE_C_SET_EXTENDED_CONFIGURATION_PARAMS:
            error=process_packet_set_extended_configuration_params(packet);
            break;

        case PACKET_CODE_C_SET_POWER:
            error=process_packet_set_power(packet);
            break;

        case PACKET_CODE_C_SET_1PPS:
            error=process_packet_set_1pps(packet);
            break;

        default:
            DEBUG("%s:unknown packet received:code=%u:size=%u:seq=%u",m_client_address_str,packet->code,packet->size,packet->seq);
            error=ENOTSUP;
            break;
    }

    ack_packet.size=PACKET_SIZE(s_ack);
    ack_packet.code=PACKET_CODE_S_ACK;
    ack_packet.s_ack.code=packet->code;
    ack_packet.s_ack.error=error;
    ack_packet.s_ack.seq=packet->seq;

    send(ack_packet);
}
